/*
 --------------------------------------------------------------------------
 ITU-T G.720.1 Source Code
 Software Release 1.1a (2010-10)
 (C) 2006 VoiceAge Corporation, Nokia Corp., Motorola Inc., Ericsson AB, and Texas Instruments Incorporated. 
 (C) 2009 Huawei Technologies Co., LTD. 
 --------------------------------------------------------------------------
*/


/*
 * wmops.c
 *
 * Cycle counting library routines for WMOPS
 * (weighted millions of operations per second)
 * estimation - adapted for G.720.1
 *
 */


#ifndef HIDE_LABEL
static const char _rev[]={"$Id: wmops.c,v 1.1 2008/11/26 12:00:41 eerikno Exp $"};
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "wmops.h"

#ifdef DRAM
/*
 * Dynamic RAM estimation tool
 *
 * Huawei Technologies Co., Ltd.
 *
 * Usage:
 *
 *   input:
 *           flag
 *               FRAME_RESET : reset the counters for frame DRAM tracking
 *               FRAME_COUNT, : actual DRAM counting
 *               DRAM_REPORT : report the maximum DRAM among the processed frames
 *           c
 *               positive value at the entry with DRAM declaration
 *               negative at the exit
 *         
 *   output: max value of the DRAM at the time been called
 *
 */
int accDRAM(int flag,int c)
{ 
  static int RAM    = 0;
  static int RAMmax = 0;
  switch(flag)
  { case FRAME_RESET:
        RAM = 0;
        break;
        
    case FRAME_COUNT:
        RAM +=c;
        if(RAM>RAMmax) 
            RAMmax=RAM;
        break;

    case DRAM_REPORT:
        break;
  }
  return RAMmax;

}
#endif /* HW_DRAM */


#if (ENABLE_WMOPS|ENABLE_TREE)

#define MAX_RECORDS     128
#define MAX_CHAR        64
#define MAX_STACK       64

#define DOUBLE_MAX  0x80000000;
struct wmops_record
{
  char   label[MAX_CHAR];
  long   call_number;
  long   update_cnt;                 /// added -- JPA 
  int    call_tree[MAX_RECORDS];
  double start_selfcnt;
  double current_selfcnt;
  double max_selfcnt;
  double min_selfcnt;
  double tot_selfcnt;
  double start_cnt;     /* The following take into account the decendants */
  double current_cnt;
  double max_cnt;
  double min_cnt;
  double tot_cnt;
};

/* globals */
static struct wmops_record wmops[MAX_RECORDS];
static int stack[MAX_STACK];
static int sptr;
static int num_records;
static int current_record;
//static int update_cnt;
static long update_cnt;

static double start_cnt;
static double max_cnt;
static double min_cnt;


#if ENABLE_WMOPS
static double inst_cnt_wc[NUM_INST];
static long   fnum_cnt_wc;             /// added -- JPA
#endif

/* These are the only global variables needing external visibility */
double ops_cnt; 
double prom_cnt;
int    ignore_depth;
#if ENABLE_WMOPS
double inst_cnt[NUM_INST];
#endif



void reset_wmops (void)
{
  int i, j;

  for (i = 0; i < MAX_RECORDS; i++)
    {
      wmops[i].label[0] = (char)NULL;
      wmops[i].call_number = 0;
      wmops[i].update_cnt = 0;             /// added -- JPA 
      for (j = 0; j < MAX_RECORDS; j++)
        {
          wmops[i].call_tree[j] = -1;
        }
      wmops[i].start_selfcnt = 0.0;
      wmops[i].current_selfcnt = 0.0;
      wmops[i].max_selfcnt = 0.0;
      wmops[i].min_selfcnt = DOUBLE_MAX;
      wmops[i].tot_selfcnt = 0.0;
      wmops[i].start_cnt = 0.0;
      wmops[i].current_cnt = 0.0;
      wmops[i].max_cnt = 0.0;
      wmops[i].min_cnt = DOUBLE_MAX;
      wmops[i].tot_cnt = 0.0;
    }

  for (i = 0; i < MAX_STACK; i++)
    {
      stack[i] = -1;
    }
  sptr = 0;
  num_records = 0;
  current_record = -1;
  update_cnt = 0;

  max_cnt = 0.0;
  min_cnt = DOUBLE_MAX;
  start_cnt = 0.0;
  ops_cnt = 0.0;
}


void push_wmops (const char *label)
{
  int new_flag;
  int i, j;

  if (ignore_depth > 0)
  {
    ignore_depth++;
  }
  else
  {
  /* Check if new function record label */
  new_flag = 1;
  for (i = 0; i < num_records; i++)
    {
      if (strcmp(wmops[i].label, label) == 0)
        {
          new_flag = 0;
          break;
        }
    }

  /* Configure new record */
  if (new_flag)
    {
      if (num_records >= MAX_RECORDS)
        {
          fprintf (stderr, "push_wmops(): exceeded MAX_RECORDS count.\n\n");
          exit(-1);
        }
      strcpy (wmops[i].label, label);
      num_records++;
    }

  /* Push current context onto stack */
  if (current_record >= 0)
    {
      if (sptr >= MAX_STACK)
        {
          fprintf (stderr, "\r push_wmops(): stack exceeded, try inreasing MAX_STACK\n");
          exit (-1);
        }
      stack[sptr++] = current_record;

      /* accumulate op counts */
      wmops[current_record].current_selfcnt += ops_cnt - wmops[current_record].start_selfcnt;

      /* update call tree */
      for (j = 0; j < MAX_RECORDS; j++)
        {
          if (wmops[i].call_tree[j] == current_record)
            {
              break;
            }
          else if (wmops[i].call_tree[j] == -1)
            {
              wmops[i].call_tree[j] = current_record;
              break;
            }
        }
    }

  /* init current record */
  current_record = i;
  wmops[current_record].start_selfcnt = ops_cnt;
  wmops[current_record].start_cnt = ops_cnt;
  wmops[current_record].call_number++;

  }
}


void pop_wmops (void)
{

  if (ignore_depth > 0)
  {
    if (ignore_depth == 1)
    {
      fprintf(stderr,"pop_wmops(): IGNORE() must be followed by push_wmops(). Exiting...\n");
      exit(-1);
    }
    if (ignore_depth == 2)
    {
      /* Resume WMOPS counting */
      ignore_depth = 0;
    }
    else
    {
      ignore_depth--;
    }
  }
  else
  {
  /* update count of current record */
  wmops[current_record].current_selfcnt += ops_cnt - wmops[current_record].start_selfcnt;
  wmops[current_record].current_cnt += ops_cnt - wmops[current_record].start_cnt;

  /* Get back previous context from stack */
  if (sptr > 0)
    {
      current_record = stack[--sptr];
      wmops[current_record].start_selfcnt = ops_cnt;
    }
  else
    {
      current_record = -1;
    }
  }
}


void update_wmops (void)
{
  int i;
  double current_cnt;

  if (sptr != 0)
    {
      fprintf (stderr, "update_wmops(): Stack must be empty!\n");
      exit (-1);
    }
  if (ignore_depth != 0)
    {
      fprintf (stderr, "update_wmops(): Missing push_wmops/pop_wmops after IGNORE(). Exiting...\n");
      exit (-1);
    }
  for (i = 0; i < num_records; i++)
    {
      wmops[i].tot_selfcnt += wmops[i].current_selfcnt;
      wmops[i].tot_cnt += wmops[i].current_cnt;

      if (wmops[i].current_selfcnt > 0)    //// Added JPA -- 11/20/2004
	{
	  if (wmops[i].current_selfcnt > wmops[i].max_selfcnt)
	    {
	      wmops[i].max_selfcnt = wmops[i].current_selfcnt;
	    }
	  if (wmops[i].current_selfcnt < wmops[i].min_selfcnt)
	    {
	      wmops[i].min_selfcnt = wmops[i].current_selfcnt;
	    }
	}
      wmops[i].current_selfcnt = 0;

      if (wmops[i].current_cnt > 0)    //// Added JPA -- 11/20/2004
	{
	  if (wmops[i].current_cnt > wmops[i].max_cnt)
	    {
	      wmops[i].max_cnt = wmops[i].current_cnt;
	    }
	  if (wmops[i].current_cnt < wmops[i].min_cnt)
	    {
	      wmops[i].min_cnt = wmops[i].current_cnt;
	    }
	  wmops[i].update_cnt++;             /// added -- JPA 
	}
      wmops[i].current_cnt = 0;
    }

  current_cnt = ops_cnt - start_cnt;
  if (current_cnt > max_cnt)
    {
      max_cnt = current_cnt;

#if ENABLE_WMOPS
      for (i = 0; i < NUM_INST; i++)
        {
          inst_cnt_wc[i] = inst_cnt[i];
        }
      fnum_cnt_wc = update_cnt+1;             /// added -- JPA 
#endif

    }

  if (current_cnt < min_cnt)
    {
      min_cnt = current_cnt;
    }

#if ENABLE_WMOPS
  for (i = 0; i < NUM_INST; i++)
    {
      inst_cnt[i] = 0.0;
    }
#endif

  start_cnt = ops_cnt;
  update_cnt++;

}


/* Real-time relationships */
#define FRAMES_PER_SECOND 100.0
#define MILLION_CYCLES     1e6
#define FAC               (FRAMES_PER_SECOND/MILLION_CYCLES)

void print_wmops (void)
{
  int i;

#if ENABLE_WMOPS

  char *sfmts = "%20s %8s %8s %7s %7s\n";
  char *dfmts = "%20s %8.2f %8.3f %7.3f %7.3f\n";
  char *sfmt =  "%20s %8s %8s %7s %7s  %7s %7s %7s\n";
  char *dfmt =  "%20s %8.2f %8.3f %7.3f %7.3f  %7.3f %7.3f %7.3f\n";

#endif

#if ENABLE_TREE

  int j;
  char *sfmtt =  "%20s %4s %15s\n";
  char *dfmtt =  "%20s %4d  ";

#endif

#if ENABLE_WMOPS

  fprintf (stdout, "\nProgram Memory Analysis: %12.0f words\n", prom_cnt);
  //fprintf (stdout, "\nInstruction Type Analysis (for worst case frame):\n\n");
  fprintf (stdout, "\nInstruction Type Analysis (for worst case frame number %ld):\n\n", fnum_cnt_wc);   /// added -- JPA
  for (i = 0; i < NUM_INST; i++)
    {
      switch ((enum instructions)i)
        {
        case _ADD:       fprintf (stdout, "\tAdds:          %12.1f\n", inst_cnt_wc[i]);  break;
        case _ABS:       fprintf (stdout, "\tAbsolutes:     %12.1f\n", inst_cnt_wc[i]);  break;
	case _MULT:      fprintf (stdout, "\tMultiplies:    %12.1f\n", inst_cnt_wc[i]);  break;
        case _MAC:       fprintf (stdout, "\tMACs:          %12.1f\n", inst_cnt_wc[i]);  break;
        case _MOVE:      fprintf (stdout, "\tMoves:         %12.1f\n", inst_cnt_wc[i]);  break;
        case _STORE:     fprintf (stdout, "\tStores:        %12.1f\n", inst_cnt_wc[i]);  break;
        case _LOGIC:     fprintf (stdout, "\tLogicals:      %12.1f\n", inst_cnt_wc[i]);  break;
        case _SHIFT:     fprintf (stdout, "\tShifts:        %12.1f\n", inst_cnt_wc[i]);  break;
        case _BRANCH:    fprintf (stdout, "\tBranches:      %12.1f\n", inst_cnt_wc[i]);  break;
        case _DIV:       fprintf (stdout, "\tDivisions:     %12.1f\n", inst_cnt_wc[i]);  break;
        case _SQRT:      fprintf (stdout, "\tSquare Root:   %12.1f\n", inst_cnt_wc[i]);  break;
        case _TRANS:     fprintf (stdout, "\tTrans:         %12.1f\n", inst_cnt_wc[i]);  break;
        case _FUNC:      fprintf (stdout, "\tFunc Call:     %12.1f\n", inst_cnt_wc[i]);  break;
        case _LOOP:      fprintf (stdout, "\tLoop Init:     %12.1f\n", inst_cnt_wc[i]);  break;
        case _INDIRECT:  fprintf (stdout, "\tIndirect Addr: %12.1f\n", inst_cnt_wc[i]);  break;
        case _PTR_INIT:  fprintf (stdout, "\tPointer Init:  %12.1f\n", inst_cnt_wc[i]);  break;
        default:         fprintf (stdout, "\tERROR: Invalid instruction type: %d\n\n", i);
        }
    }

  fprintf (stdout, "\n\nWeighted MOPS Analysis:\n");
  fprintf (stdout, "%54s  %23s\n", "|------  SELF  ------|"
                                  ,"|---  CUMULATIVE  ---|");
  fprintf (stdout, sfmt, "        routine", " calls", "  min ", "  max ", "  avg ", "  min ", "  max ", "  avg ");
  fprintf (stdout, sfmt, "---------------", "------", "------", "------", "------", "------", "------", "------");
  for (i = 0; i < num_records; i++)
    {
      fprintf (stdout, dfmt, 
                   wmops[i].label,
            (float)wmops[i].call_number/update_cnt,
               FAC*wmops[i].min_selfcnt,
               FAC*wmops[i].max_selfcnt,
               //FAC*wmops[i].tot_selfcnt/update_cnt,
               FAC*wmops[i].tot_selfcnt/wmops[i].update_cnt,             /// added -- JPA
               FAC*wmops[i].min_cnt,
               FAC*wmops[i].max_cnt,
               //FAC*wmops[i].tot_cnt/update_cnt);
               FAC*wmops[i].tot_cnt/wmops[i].update_cnt);             /// added -- JPA
    }
  fprintf (stdout, sfmts, "---------------", "------", "------", "------", "------");
  fprintf (stdout, dfmts, 
           "total",
           (float)update_cnt,
           FAC*min_cnt,
           FAC*max_cnt,
           FAC*ops_cnt/update_cnt);
  fprintf (stdout, "\n"); 

#endif

#if ENABLE_TREE

  fprintf (stdout, "\nCall Tree:\n\n");
  fprintf (stdout, sfmtt, "       function", "num", "called by:    ");
  fprintf (stdout, sfmtt, "---------------", "---", "--------------");

  for (i = 0; i < num_records; i++)
    {
      fprintf (stdout, dfmtt, wmops[i].label, i);
      for (j = 0; wmops[i].call_tree[j] != -1; j++)
        {
          if (j != 0)
            {
              fprintf (stdout, ", ");
            }
          fprintf (stdout, "%d", wmops[i].call_tree[j]);
        }
      fprintf (stdout, "\n");

    }

  fprintf (stdout, sfmtt, "---------------", "---", "--------------");
  fprintf (stdout, "\n\n");

#endif
}

#endif


